
import java.util.List;
import java.util.ArrayList;

/**
   Here is a summary of the methods in this class.
      int size()
      void add(E)
      boolean search(E)
      boolean remove(E)
      E min()
      E max()
      E predecessor(E)
      E successor(E)
      List<E> inOrder()
*/
public class BinarySearchTree<E extends Comparable<E>>
{
   private E data; // if this field is null, then the tree is empty
   private BinarySearchTree<E> left;
   private BinarySearchTree<E> right;
   private int size;

   /**
      Construct an empty BinarySearchTree.
   */
   public BinarySearchTree()
   {
      this.data  = null;  // an empty binary search tree
      this.left  = null;
      this.right = null;
      this.size = 0;
   }


   /**
      Construct a BinarySearchTree from a data element and two given trees.

      This constructor is private because, in general, you cannot
      construct a BinarySearchTree from two other BinarySearchTree
      objects. This constructor is used in the removeP() method.

      @param data   root element for the new BinarySearchTree
      @param left   leftsub tree for the new BinarySearchTree
      @param right  right sub tree for the new BinarySearchTree
   */
   private BinarySearchTree(E data,
                            BinarySearchTree<E> left,
                            BinarySearchTree<E> right)
   {
      this.data  = data;
      this.left  = left;
      this.right = right;
      this.size = 1 + left.size + right.size;
   }


   /**
      Return the number of nodes in this BinarySearchTree.

      @return the number of nodes in this BinarySearchTree
   */
   public int size()  // O(1)
   {
      return this.size;
   }


   /**
      Add an element to this BinarySearchTree.

      @param data  element to place into the binary searh tree
      @throws NullPointerException
   */
   public void add(E data)  // O(lg(n)) on average, worst case O(n)
   {
      if (null == data)
         throw new NullPointerException("data cannot be null");

      if (0 == this.size) // Create a leaf node.
      {
         this.data  = data;
         this.left  = new BinarySearchTree<E>();
         this.right = new BinarySearchTree<E>();
         this.size = 1;
         return;
      }

      final int comp = data.compareTo(this.data);
      if (0 == comp)
      {
         return; // The data element is already in the search tree.
      }
      else if (0 > comp)
      {
         left.add(data);  // recursion
      }
      else
      {
         right.add(data);  // recusion
      }
      ++size;
   }


   /**
      Determine if an element of data is
      contained in this BinarySearchTree.

      @param data  element of data to search for
      @return true if the data element is contained in this BinarySearchTree
      @throws NullPointerException
   */
   public boolean search(E data)  // O(lg(n)) on average, worst case O(n)
   {
      if (null == data)
         throw new NullPointerException("data cannot be null");

      if (0 == this.size) // Empty tree, search fails.
      {
         return false;
      }

      final int comp = data.compareTo(this.data);
      if (0 == comp)
      {
         return true; // found
      }
      else if (0 > comp)
      {
         return left.search(data);  // recusion
      }
      else
      {
         return right.search(data);  // recursion
      }
   }


   /**
      Find the minimum element in thes BinarySearchTree.

      @return the minimum element in this BinarySearchTree
   */
   public E min()  // O(n)
   {
      // Find the fist mode with no left child.
      if (0 == this.left.size())
      {
         return this.data;
      }
      else
      {
         return this.left.min();  // recursion
      }
   }


   /**
      Find the maximium element in thes BinarySearchTree.

      @return the maximium element in this BinarySearchTree
   */
   public E max()  // O(n)
   {
      // Find the fist mode with no right child.
      if (0 == this.right.size())
      {
         return this.data;
      }
      else
      {
         return this.right.max();  // recursion
      }
   }


   /**
      Find the predecessor element of the givien
      data element in thes BinarySearchTree.

      @param data  element of data to find the predecessor of
      @return the predecessor element in this BinarySearchTree
      @throws NullPointerException
      @throws IllegalStateException
   */
   public E predecessor(E data)  // O(lg(n)) on average, worst case O(n)
   {
      if (null == data)
         throw new NullPointerException("data cannot be null");

      if (0 == this.size) // Empty tree.
         throw new IllegalStateException("empty search tree");

      BinarySearchTree<E> current = this;
      BinarySearchTree<E> pred = null;
      while ( 0 != current.size() ) // while not empty
      {
         if ( data.compareTo(current.data) > 0 ) // If current node is less,
         {                                       // then it is a potential predecessor.
            pred = current;
            current = current.right;
         }
         else // data >= current.data
         {
            current = current.left;
         }
      }
      if (null != pred) return pred.data;
      else throw new IllegalStateException("no predecessor to " + data);
   }


   /**
      Find the successor element of the givien
      data element in thes BinarySearchTree.

      @param data  element of data to find the successor of
      @return the successor element in this BinarySearchTree
      @throws NullPointerException
      @throws IllegalStateException
   */
   public E successor(E data)  // O(lg(n)) on average, worst case O(n)
   {
      if (null == data)
         throw new NullPointerException("data cannot be null");

      if (0 == this.size) // Empty tree.
         throw new IllegalStateException("empty search tree");

      BinarySearchTree<E> current = this;
      BinarySearchTree<E> succ = null;
      while ( 0 != current.size() ) // while not empty
      {
         if ( data.compareTo(current.data) < 0 ) // If current node is greater,
         {                                       // then it is a potential successor.
            succ = current;
            current = current.left;
         }
         else // data >= current.data
         {
            current = current.right;
         }
      }
      if (null != succ) return succ.data;
      else throw new IllegalStateException("no successor to " + data);
   }


   /**
      Return a new BinarySearchTree with the given data
      element reoved from it. If the data is not in this
      BinarySearchTree, then return this tree unchanged.

      This method is written in the style of a persistent data structure.
         https://en.wikipedia.org/wiki/Persistent_data_structure

      @param data  element to remove if its in this BinarySearchTree
      @return a new BinarySearchTree with data removed
   */
   public BinarySearchTree<E> removeP(E data)  // O(lg(n)) on average, worst case O(n)
   {
      if ( ! this.search(data) ) // data is not an element of this tree
      {
         return this;
      }
      else if ( ! data.equals(this.data) ) // data is not the root of this tree
      {
         return new BinarySearchTree<E>(this.data,
                                        this.left.removeP(data),   // recursion
                                        this.right.removeP(data)); // recursion
      }
      else // remove the root node
      {
         if (1 == this.size) // the root is a leaf
         {
            return new BinarySearchTree<E>();
         }
         else if (0 == this.left.size)
         {
            return this.right;
         }
         else if (0 == this.right.size)
         {
            return this.left;
         }
         else
         {
            // Find the successor of the root element.
            final E successsor = this.right.min();
            return new BinarySearchTree<E>(successsor,
                                           this.left,
                                           this.right.removeP(successsor)); // recursion
         }
      }
   }


   /**
      Remove the given data element from this BinarySearchTree.

      This method mutates (modifies) the current tree.

      @param data  element to remove if its in this BinarySearchTree
      @return true if the data element is in the tree and was removed
   */
   public boolean remove(E data)  // O(lg(n)) on average, worst case O(n)
   {
      if ( ! this.search(data) ) // data is not an element of this tree
      {
         return false;
      }
      else if ( ! data.equals(this.data) ) // data is not the root of this tree
      {
         this.left.remove(data);   // recursion
         this.right.remove(data);  // recursion
         --this.size;
         return true;
      }
      else // remove the root node
      {
         if (1 == this.size) // the root is a leaf
         {
            this.data = null;  // create and empty tree
            this.left = null;
            this.right = null;
         }
         else if (0 == this.left.size)
         {
            this.data  = this.right.data;
            this.left  = this.right.left;
            this.right = this.right.right;
         }
         else if (0 == this.right.size)
         {
            this.data  = this.left.data;
            this.right = this.left.right; // Be careful with the order
            this.left  = this.left.left;  // of these two assignments!
         }
         else
         {
            // Find the successor of the root element.
            final E successsor = this.right.min();
            // Remove the root's successor from the root's right sub tee.
            this.right.remove(successsor); // recursion
            // Replace the root's data element with its successor.
            this.data = successsor;
         }
         --this.size;
         return true;
      }
   }


   /**
      Return a List of the nodes from this BinarySearchTree
      in order from smallest to largest.

      Do an in-order traversal of this binary tree.

      @return a List of the elements of this BinarySearchTree in ascending order
   */
   public List<E> inOrder()  // O(n)
   {
      if (0 == this.size)
      {
         return new ArrayList<>();
      }
      else
      {
         final List<E> result = left.inOrder();
         result.add(this.data);
         result.addAll(right.inOrder());
         return result;
      }
   }


   @Override
   public String toString()
   {
      return (null == data)? "()" :
               (0 == left.size && 0 == right.size)? "" + data :
                 "(" + data + ", " + left + ", " + right + ")";
   }
}
