import java.util.Random;

/**
   This program measures the runtime of binary
   search tree searches on a sequence of growing
   binary search trees.

   First we profile searches on trees made up of
   randomly distributed numbers. The search trees
   will, on average, be well balanced and the
   search times will be O(lg(n)) on the size of
   the tree.

   Then we profile searches on trees made up of
   sequential numbers. These search trees will
   be totally unbalanced, so the search times
   will be O(n) on the size of the trees.
*/
public class BinarySearchTree_Profile
{
   final static Random generator = new Random();

   public static void main(String[] args)
   {
      int size = 1_000;    // starting size for the trees
      int iterations = 17; // how many times to increase the tree size

      System.out.println("Binary Search Trees -- average case, O( lg(n) ).");

      BinarySearchTree<Integer> bst = new BinarySearchTree<>();
      for (int i = 1; i <= iterations; ++i)
      {
         randomIntBST(bst, size, 100_000_000); // build the tree

         long startTime = System.currentTimeMillis();
         // Search the binary search tree 1,000,000 times.
         int count = 0;
         for (int j = 0; j < 1_000_000; ++j)
         {
            final int key = generator.nextInt( 100_000_000 );
            final boolean b = bst.search( key );
            if (b) ++count;
         }
         long stopTime = System.currentTimeMillis();
         final long avgCase =(stopTime - startTime);

         System.out.println(
            "For n = " + size + ", "
          + "avg time = " + avgCase
          + " (% hits = " + 100 * count / 1_000_000.0 + "%)");

         size = 2 * size; // incrrease the size for the next tree
      }


      System.out.println("Binary Search Trees -- worst case, O( n ).");

      size = 1_000;    // starting size for the trees
      iterations = 20; // how many times to increase the tree size
      for (int i = 1; i <= iterations; ++i)
      {
         BinarySearchTree<Integer> bst2 = new BinarySearchTree<>();
         for (int j = 0; j < i*size; ++j)
         {
            bst2.add( j );
         }
         long startTime = System.currentTimeMillis();
         // Search the binary search tree 10,000 times.
         int count = 0;
         for (int j = 0; j < 10_000; ++j)
         {
            final int key = generator.nextInt( 10*i*size );
            final boolean b = bst2.search( key );
            if (b) ++count;
         }
         long stopTime = System.currentTimeMillis();
         final long worstCase =(stopTime - startTime);

         System.out.println(
            "For n = " + (i*size) + ", "
          + "worst time = " + worstCase
          + " (% hits = " + 100 * count / 10_000.0 + "%)");
      }
   }


   /**
      Add random integer values to a BinarySearchTree.

      @param bst   the BinarySearchTree to insert integers into
      @param size  the number of integers to add to the BinarySearchTree
      @param n     the range of possible random values
   */
   public static void randomIntBST(BinarySearchTree<Integer> bst,
                                   int size, int n)
   {
      for (int i = 0; i < size; ++i)
      {
         bst.add( generator.nextInt(n) );
      }
   }
}
